Require Import List.
Require Import Classes.EquivDec.
Require Import LIFO.Tactics.

Set Implicit Arguments.

Fixpoint assoc (A B:Type)(eqA: EqDec A eq)(a:A)(l:list (A*B)): option B :=
  match l with 
    | nil => None 
    | cons h t => if (eqA a (fst h)) then Some (snd h) else assoc eqA a t
  end.  

Lemma assocApp: 
  forall (A B: Type)(eqA: EqDec A eq)(a:A)(l1 l2:list (A*B)),
    (forall b, ~In (a,b) l1) -> assoc eqA a (l1++l2) = assoc eqA a l2.
Proof.
  intros A B eqA a; induction l1; intros l2 H.
    auto.
    simpl; case_eq (eqA a (fst a0)); intros Heq Heq'.
      set(H':=H (snd a0)); contradict H'; destruct a0; simpl in *;
        left; rewrite Heq; reflexivity.
      apply IHl1; intros b Hb; apply (H b); right; assumption.
Qed.

Definition inFirst (A B: Type)(x:(A*B))(l:list (A*B)) :=
  exists l1 :list(A*B), 
    exists l2 :list(A*B), 
    l = l1++x::l2 /\ (forall b, ~(In ((fst x),b) l1)).

Lemma assocInFirst: 
  forall (A B:Type)(eqA: EqDec A eq)(a:A)(l:list (A*B))(b:B),
    assoc eqA a l = Some b <-> inFirst (a,b) l.
Proof.
  intros A B eqA a l; generalize dependent a; induction l; 
  split; intro H; simpl in H.
  (* Case nil *)
    (* -> *)
    discriminate.
    (* <- *)
    unfold inFirst in H; simpl in H; destruct H as [l1 H]; 
      destruct H as [l2 H]; destruct H as [H _]; simpl in H; 
        contradict H; apply app_cons_not_nil.
  (* Case cons *)
    (* -> *)
    case_eq (eqA a0 (fst a)); intros Heq Heq'; rewrite Heq' in H; simpl in H.          (* Case a0 = fst a *)
      inversion H; rewrite Heq in *; destruct a; simpl; unfold inFirst; simpl.
        exists (@nil (A*B)); exists l; split; auto.
      (* Case a0<> fst a *)
      assert(inFirst (a0,b) l) as Hl by (apply IHl; assumption);
        unfold inFirst in *; destruct Hl as [l1 Hl]; destruct Hl as [l2 Hl]; 
          destruct Hl as [Hl1 Hl2]; exists (cons a l1); exists l2; split.
            rewrite <- app_comm_cons; f_equal; assumption.
            intros b0 H0; simpl in H0; destruct H0.
              destruct a; inversion H0; subst;  elim Heq; simpl; apply eq_refl.
              apply (Hl2 b0); assumption.
    (* <- *)
    unfold inFirst in H; 
      destruct H as [l1 H]; destruct H as [l2 H]; destruct H as [H1 H2];
        rewrite H1; rewrite assocApp; try assumption; simpl;
          case_eq(eqA a0 a0); intros H; auto; elim H; apply eq_refl.
Qed.


Definition NoDupKey K A  (l: (list (K*A))) := NoDup (map fst l).
Implicit Arguments NoDupKey [K A].      


Lemma in_inFirst :
  forall K A (a:A) (l:list (K*A)),            
    NoDupKey l ->
    In a  ((map snd) l) ->             
    exists k,   inFirst (k, a) l.
Proof.
  intros K A0 a l H H0.          
  unfold inFirst.
  apply in_map_iff in H0.
  destruct H0 as [[k v'] [v_v' in_kv_l]].
  simpl in v_v'; subst.
  induction l.
  simpl in in_kv_l; contradiction.
  copy H as no_dup_l. 
  unfold NoDupKey in *. 
  simpl in no_dup_l. 
  change (fst a0 :: map fst l) with (nil++fst a0 :: map fst l) in no_dup_l.
  apply NoDup_remove_1 in no_dup_l. simpl in no_dup_l.
  simpl in in_kv_l; destruct in_kv_l.
  destruct a0;injection H0.
  intros H1 H2; subst.          
  exists k; exists nil; exists l; split ;[reflexivity|simpl;auto].                
  
  destruct (IHl no_dup_l H0) as [k' [ l1 [l2 [ l_eq not_in_l1]]]].
  subst.          
  change  (a0 :: l1 ++ (k', a) :: l2) with  ( (a0 :: l1) ++ (k', a) :: l2) in H.
  rewrite map_app in H.
  change  (map fst (a0 :: l1) ++ map fst ((k', a) :: l2)) with  (map fst (a0 :: l1) ++ fst (k', a) :: map fst ( l2)) in H.
  apply NoDup_remove_2 in H.
  exists k'; exists (a0::l1); exists l2; split ;[reflexivity|simpl;auto].                
  intros b.            
  simpl in *. 
  intro not_in_k'.   
  destruct not_in_k' as [not_in_k'|not_in_k'] .
  destruct a0.
  destruct H.
  left.           rewrite not_in_k'. reflexivity.
  apply Decidable.not_or in H.
  destruct H as [fst_a0_diff_k  not_in_k'_map_fst].
  elim not_in_k'_map_fst.          
  apply in_app_iff.
  left.
  apply in_map_iff.
  exists (k',b); simpl; auto.
Qed.
Implicit Arguments in_inFirst [K A ].
