Require Import List.
Require Import Sorting.Permutation.

Require Import LIFO.List.Fold.
Require Import LIFO.List.Length.
Require Import LIFO.List.App.
Require Import LIFO.Algebra.AlgebraClasses.
Set Implicit Arguments.

(** * definition of flatten 
   flatten append together the lists contain in a list*)
Definition flatten (A:Type)(l:list(list A)) : list A := 
  List.fold_right (List.app (A:=A)) nil l.

Lemma flattenApp:
  forall (A:Type)(l1 l2:list(list A)), flatten (l1++l2) = (flatten l1)++(flatten l2).
Proof.
  intros A l1 l2.
  induction l1; trivial.
  simpl; rewrite IHl1; intuition.
Qed.

(*  if [ll] contains a non empty lists then flatten is not empty *)
Lemma flattenNotNil:
  forall (A:Type)(ll:list(list A))(H:forall l, In l ll -> l<>nil),
    ll<>nil -> flatten ll<> nil.
Proof.
  induction ll; intros H H0. 
    elim H0; trivial.
    simpl; destruct a. 
      elim (H nil); intuition.
      discriminate.
Qed.

Hint Unfold flatten.


(** length of flatten is the sum of the length of sub-lists  *)
Lemma flatten_length: forall (A:Type)(l:list(list A)),
  List.length (flatten l) = List.fold_right plus 0 (List.map (@List.length A) l).
Proof.
  induction l;
    trivial;
      simpl; unfold flatten; simpl; autorewrite with length; trivial.
Qed.

Hint Rewrite flatten_length : length.

(** if [a] is in [flatten ll] then it is in one of the lists in ll  *)
Lemma inFlatten: forall (A:Type)(ll:list(list A))(a:A), 
  In a (flatten ll) <-> exists l:list A, In l ll /\ In a l.
Proof.
  intros A ll a; split; intro H; unfold flatten; induction ll as [| head tail ].
    contradict H; trivial.
    simpl in H. rewrite in_app_iff in H; destruct H.
      exists head; intuition.
      destruct(IHtail H) as [l Hl]; exists l; intuition.
      destruct H; intuition. 
      simpl; apply in_app_iff; destruct H as [l Hl]; destruct Hl as [Hl1 Hl2]; 
        simpl in Hl1; destruct Hl1; 
          [ left; subst; auto | 
            right; apply IHtail; eauto].
Qed.

(** If [l] and [l'] are permutation then flattening them leads to a permutation  *)
Lemma Permutation_flatten : forall A (l l': list (list A)), 
  Permutation l l' -> Permutation (flatten l) (flatten l').
Proof.
  intros A0. 
  apply (Permutation_ind ); auto.
  intros x l l' H H0.        
  simpl.
  apply Permutation_app_head; assumption.
  intros x y l.
  simpl.
  rewrite <- (AlgebraClasses.associative  y x).
  rewrite <- (AlgebraClasses.associative  x y).
  apply Permutation_app_tail.
  apply Permutation_app_comm.
  intros l l' l'' H H0 H1 H2.
  
  rewrite H0.
  rewrite H2.
  reflexivity.
Qed.

Lemma flattenFoldLeftApp:
  forall (A:Type)(ll:list(list A)),
    flatten ll = fold_left (fun x y => x++y) ll nil.
Proof.
  induction ll. 
  reflexivity.    
  simpl.
  rewrite IHll.
 rewrite <- fold_left_app_acc.
 reflexivity.
Qed.

Implicit Arguments flatten [A].
Implicit Arguments flattenApp [A].
Implicit Arguments flattenNotNil [A].
Implicit Arguments flatten_length [A].
Implicit Arguments inFlatten [A].
Implicit Arguments Permutation_flatten [A].
Implicit Arguments flattenFoldLeftApp [A].
      
    
